/*----------------------------------------------------------------------------

  Copyright (c) 2017 rafael grompone von gioi <grompone@gmail.com>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.

  ----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "iio.h"

/*----------------------------------------------------------------------------*/
#ifndef FALSE
#define FALSE 0
#endif /* !FALSE */

#ifndef TRUE
#define TRUE 1
#endif /* !TRUE */

/*----------------------------------------------------------------------------*/
/*
   Fatal error, print a message to standard-error output and exit.
 */
static void error(char * msg)
{
  fprintf(stderr,"Error: %s\n",msg);
  exit(EXIT_FAILURE);
}

/*----------------------------------------------------------------------------*/
/** Memory allocation, print an error and exit if fail.
 */
static void * xmalloc(size_t size)
{
  void * p;
  if( size == 0 ) error("xmalloc: zero size");
  p = malloc(size);
  if( p == NULL ) error("xmalloc: out of memory");
  return p;
}

/*----------------------------------------------------------------------------*/
/*
   Computes the logarithm of NFA to base 10.

   NFA = NT.b(n,k,p)
   the return value is log10(NFA)

   n,k,p - binomial parameters.
   logNT - logarithm of Number of Tests
 */
#define TABSIZE 100000
static double log_nfa(int n, int k, double p, double logNT)
{
  static double inv[TABSIZE];   /* table to keep computed inverse values */
  double tolerance = 0.1;       /* an error of 10% in the result is accepted */
  double log1term,term,bin_term,mult_term,bin_tail,err;
  double p_term = p / (1.0-p);
  int i;

  if( n<0 || k<0 || k>n || p<0.0 || p>1.0 )
    error("wrong n, k or p values in nfa()");

  if( n==0 || k==0 ) return logNT;
  if( n==k ) return logNT + (double)n * log10(p);

  log1term = lgamma((double)n+1.0) - lgamma((double)k+1.0)
           - lgamma((double)(n-k)+1.0)
           + (double)k * log(p) + (double)(n-k) * log(1.0-p);

  term = exp(log1term);
  if( term == 0.0 )                        /* the first term is almost zero */
    {
      if( (double)k > (double)n * p )      /* at begining or end of the tail? */
        return log1term / M_LN10 + logNT;  /* end: use just the first term */
      else
        return logNT;                      /* begin: the tail is roughly 1 */
    }

  bin_tail = term;
  for(i=k+1;i<=n;i++)
    {
      bin_term = (double)(n-i+1) * ( i<TABSIZE ?
                   (inv[i] ? inv[i] : (inv[i]=1.0/(double)i)) : 1.0/(double)i );
      mult_term = bin_term * p_term;
      term *= mult_term;
      bin_tail += term;
      if(bin_term<1.0)
        {
          /* when bin_term<1 then mult_term_j<mult_term_i for j>i.
             then, the error on the binomial tail when truncated at
             the i term can be bounded by a geometric serie of form
             term_i * sum mult_term_i^j.                            */
          err = term * ( ( 1.0 - pow( mult_term, (double)(n-i+1) ) ) /
                         (1.0-mult_term) - 1.0 );

          /* one wants an error at most of tolerance*final_result, or:
             tolerance * abs(-log10(bin_tail)-logNT).
             now, the error that can be accepted on bin_tail is
             given by tolerance*final_result divided by the derivative
             of -log10(x) when x=bin_tail. that is:
             tolerance * abs(-log10(bin_tail)-logNT) / (1/bin_tail)
             finally, we truncate the tail if the error is less than:
             tolerance * abs(-log10(bin_tail)-logNT) * bin_tail        */
          if( err < tolerance * fabs(-log10(bin_tail)-logNT) * bin_tail ) break;
        }
    }
  return log10(bin_tail) + logNT;
}

/*----------------------------------------------------------------------------*/
/*                                    Main                                    */
/*----------------------------------------------------------------------------*/
int main(int argc, char ** argv)
{
  double * input;
  double * image;
  double * cross_diff;
  double * cumulated;
  int X,Y,C;
  int x,y,c,i,j,k,Bx,By;
  int xmax = -1;
  int ymax = -1;
  double max,lnfa;
  int N;

  /* read input */
  if( argc != 5 )
    error("use: jpegblocks <block_size> <image> <cross_diff_out> "
                "<cumulated_out> ");
  N = atoi(argv[1]);
  input = iio_read_image_double_split(argv[2], &X, &Y, &C);

  /* get memory */
  image = (double *) xmalloc( X * Y * sizeof(double) );
  cross_diff = (double *) xmalloc( X * Y * sizeof(double) );
  cumulated = (double *) xmalloc( N * N * sizeof(double) );

  /* generate a gray level image from the input
     (this is useful when the input has many channels) */
  for(x=0; x<X; x++)
    for(y=0; y<Y; y++)
      {
        image[x+y*X] = 0.0;
        for(c=0; c<C; c++)
          image[x+y*X] += input[x+y*X+c*X*Y];
      }

  /* compute absolute value of cross-difference */
  for(x=0; x<(X-1); x++)
    for(y=0; y<(Y-1); y++)
      cross_diff[x+y*X] = fabs( image[x+y*X] + image[x+1+(y+1)*X]
                              - image[x+1+y*X] - image[x+(y+1)*X] );

  /* compute cumulated cross-difference modulo block position */
  for(x=0; x<N; x++)         /* initialization */
    for(y=0; y<N; y++)
      cumulated[x+y*N] = 0.0;
  Bx = (X-1) / N; /* number of blocks to consider */
  By = (Y-1) / N; /* (the last column and row don't have a cross-difference
                      value beacause forward difference are used. only full
                      blocks without those row and column are considered) */
  for(i=0; i<Bx; i++)        /* cumulate values */
    for(j=0; j<By; j++)
      for(x=0; x<N; x++)
        for(y=0; y<N; y++)
          cumulated[x+y*N] += cross_diff[ N*i+x + (N*j+y) * X ];

  /* find x of maximal column */
  max = 0.0;
  for(x=0; x<N; x++)
    {
      double v = 0.0;
      for(y=0; y<N; y++) v += cumulated[x+y*N];
      if( v > max )
        {
          max = v;
          xmax = x;
        }
    }

  /* find y of maximal row */
  max = 0.0;
  for(y=0; y<N; y++)
    {
      double v = 0.0;
      for(x=0; x<N; x++) v += cumulated[x+y*N];
      if( v > max )
        {
          max = v;
          ymax = y;
        }
    }

  /* compute number of favorable cases for maximal position */
  k = 0;
  for(i=0; i<N; i++)
    {
      if( cumulated[i+ymax*N] > cumulated[ i + ((ymax-1)%N) * N] ) ++k;
      if( cumulated[i+ymax*N] > cumulated[ i + ((ymax+1)%N) * N] ) ++k;
      if( cumulated[xmax+i*N] > cumulated[ (xmax-1)%N  +  i * N] ) ++k;
      if( cumulated[xmax+i*N] > cumulated[ (xmax+1)%N  +  i * N] ) ++k;
    }

  /* compute log(NFA) of xmax,ymax being the lower-right position of blocks */
  lnfa = log_nfa( 4*N, k, 0.5, log10( (double) N * N ) );

  /* print output */
  printf("Cross-difference cumulated matrix:\n");
  printf("     ");
  for(x=0; x<N; x++)
    if( x == xmax ) printf("    xmax ");
    else            printf("         ");
  printf("\n");
  for(y=0; y<N; y++)
    {
      if( y == ymax ) printf("ymax ");
      else            printf("     ");
      for(x=0; x<N; x++) printf("%8g ",cumulated[x+y*N]);
      printf("\n");
    }
  printf("\nk = %d\n",k);
  printf("Block origin %d %d log(NFA) %g  ->  %s\n",
         (xmax+1)%N, (ymax+1)%N, lnfa,
         lnfa < 0.0 ? "JPEG grid detected" : "JPEG grid NOT detected" );

  /* normalize and save cross_diff */
  for(x=0; x<X; x++)
    for(y=0; y<Y; y++)
      cross_diff[x+y*X] = cross_diff[x+y*X] < 25 ? 10*cross_diff[x+y*X] : 255.0;
  iio_write_image_double_split(argv[3],cross_diff,X,Y,1);

  /* normalize and save cumulated */
  max = 0.0;
  for(x=0; x<N; x++)
    for(y=0; y<N; y++)
      if( cumulated[x+y*N] > max ) max = cumulated[x+y*N];
  for(x=0; x<N; x++)
    for(y=0; y<N; y++)
      cumulated[x+y*N] *= 255.0 / max;
  iio_write_image_double_split(argv[4],cumulated,N,N,1);

  /* free memory */
  free( (void *) input );
  free( (void *) image );
  free( (void *) cross_diff );
  free( (void *) cumulated );

  return EXIT_SUCCESS;
}
/*----------------------------------------------------------------------------*/
